package org.pasoa.dommap;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Provides convience methods for writing a org.w3c.dom.document
 * to a string.
 * 
 * @author sm
 *
 */
public class DOMWriter {
    private Writer  out;
    private boolean format;
    private boolean _namespaceMode;
    
    public DOMWriter (Writer output) {
        out            = new BufferedWriter (output);
        format         = false;
        _namespaceMode = false;
    }
    
    public DOMWriter (OutputStream output) {
        out    = new OutputStreamWriter (output);
        format = false;
        _namespaceMode = false;
    }
    
    private DOMWriter (Writer output, boolean namespaceMode) {
        _namespaceMode = namespaceMode;
    }
    
    private String getQualifiedName (Node node) {
        String prefix = node.getPrefix ();
        
        if (_namespaceMode) {
            if (node.getNamespaceURI () == null) {
                return "null";
            } else {
                return node.getNamespaceURI ();
            }
        }
        if (node.getNodeName ().equals ("xmlns")) {
            return "xmlns";
        }
        if (prefix == null) {
            if (node.getNamespaceURI () == null) {
                return node.getNodeName ();
            } else {
                prefix = "ns";
            }
        }
        
        return prefix + ":" + node.getLocalName ();
    }
    
    private boolean hasChildren (Element node, boolean onlyElementsCount) {
        NodeList children = node.getChildNodes ();
        boolean  found    = false;
        Node     child;
        
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (child instanceof Element) {
                found = true;
            }
            if (!onlyElementsCount && child instanceof Text) {
                found = true;
            }
        }
        
        return found;
    }
    
    private void indent (int indent) throws IOException {
        while (indent > 0) {
            out.write (" ");
            indent -= 1;
        }
    }
    
    public void setFormatted (boolean isFormatted) {
        format = isFormatted;
    }
    
    private String substitute (String text, String from, String to) {
        int index = text.indexOf (from);
        
        while (index >= 0) {
            text  = text.substring (0, index) + to + text.substring (index + from.length ());
            index = text.indexOf (from, index + to.length ());
        }
        
        return text;
    }
    
    public void write (Attr node) throws IOException {
        String qname = getQualifiedName (node);
        
        out.write (" " + qname + " = \"" + node.getNodeValue () + "\"");
        out.flush ();
    }
    
    /**
     * Write a document out specifying which namespaces have already been declared
     * and the initial indent of elements.
     */
    public void write (Document node, Map declared, int indent) throws IOException {
        write (node.getDocumentElement (), declared, indent);
        out.flush ();
    }
    
    /**
     * Convenience method for writing a document with no namespaces already declared
     * and 0 initial indent.
     */
    public void write (Document node) throws IOException {
        write (node, new HashMap (), 0);
    }
    
    public void write (Element node) throws IOException {
        write (node, new HashMap (), 0);
    }
    
    public void write (Element node, Map declared, int indent) throws IOException {
        String qname = getQualifiedName (node);
        
        if (format) {
            indent (indent);
        }
        out.write ("<" + qname);
        writeAttributes (node, declared);
        if (!hasChildren (node, false)) {
            out.write ("/>");
            if (format) {
                out.write ("\n");
            }
        } else {
            out.write (">");
            if (format && hasChildren (node, true)) {
                out.write ("\n");
            }
            writeChildren (node, declared, indent + 1);
            if (format && hasChildren (node, true)) {
                indent (indent);
            }
            out.write ("</" + qname + ">");
            if (format) {
                out.write ("\n");
            }
        }
        out.flush ();
    }
    
    public void write (Text node) throws IOException {
        String text = substitute (substitute (substitute (substitute (substitute (node.getNodeValue (), "&", "&amp;"), "<", "&lt;"), ">", "&gt;"), "'", "&apos;"), "\"", "&quot;");
        
        out.write (text);
        out.flush ();
    }
    
    private void writeAttributes (Element node, Map declared) throws IOException {
        NamedNodeMap children = node.getAttributes ();
        Node         child;
        
        if (node instanceof SimpleElement) {
            writeNamespace (node, declared);
        }
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (node instanceof SimpleElement) {
                writeNamespace (child, declared);
            }
            write ((Attr) child);
        }
    }
    
    private void writeChildren (Element node, Map declared, int indent) throws IOException {
        NodeList children = node.getChildNodes ();
        Node     child;
        
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (child instanceof Element) {
                write ((Element) child, new HashMap (declared), indent);
            }
            if (child instanceof Text) {
                write ((Text) child);
            }
        }
    }
    
    private void writeNamespace (Node node, Map declared) throws IOException {
        String prefix = node.getPrefix ();
        
        if (prefix == null && node.getNamespaceURI () == null) {
            return;
        }
        if (prefix == null) {
            prefix = "ns";
        }
        
        if (!declared.containsKey (prefix) || !declared.get (prefix).equals (node.getNamespaceURI ())) {
            out.write (" xmlns:" + prefix + " = \"" + node.getNamespaceURI () + "\"");
            declared.put (prefix, node.getNamespaceURI ());
        }
    }
    
    public static String writeNamespaces (Node toWrite){
        StringWriter output = new StringWriter ();
        DOMWriter    writer = new DOMWriter (output, true);
        
        try {
            if (toWrite == null) {
                throw new RuntimeException ("Trying to write null Node");
            }
            writer.setFormatted (false);
            if (toWrite instanceof Attr) {
                writer.write ((Attr) toWrite);
            }
            if (toWrite instanceof Document) {
                writer.write ((Document) toWrite);
            }
            if (toWrite instanceof Element) {
                writer.write ((Element) toWrite);
            }
            if (toWrite instanceof Text) {
                writer.write ((Text) toWrite);
            }
            output.close ();
        } catch (IOException shouldNeverHappen) {
            throw new RuntimeException ("IOException with StringWriter in DOMWriter: " + shouldNeverHappen);
        }
        
        return output.getBuffer ().toString ();
    }
    
    public static String writeToString (Node toWrite) {
        return writeToString (toWrite, false);
    }
    
    public static String writeToString (Node toWrite, boolean format){
        StringWriter output = new StringWriter ();
        DOMWriter    writer = new DOMWriter (output);
        
        try {
            if (toWrite == null) {
                throw new RuntimeException ("Trying to write null Node");
            }
            writer.setFormatted (format);
            if (toWrite instanceof Attr) {
                writer.write ((Attr) toWrite);
            }
            if (toWrite instanceof Document) {
                writer.write ((Document) toWrite);
            }
            if (toWrite instanceof Element) {
                writer.write ((Element) toWrite);
            }
            if (toWrite instanceof Text) {
                writer.write ((Text) toWrite);
            }
            output.close ();
        } catch (IOException shouldNeverHappen) {
            throw new RuntimeException ("IOException with StringWriter in DOMWriter: " + shouldNeverHappen);
        }
        
        return output.getBuffer ().toString ();
    }
}
