package org.pasoa.dommap;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public abstract class SimpleNode extends NonImplementedNode {
    private List     children;
    private String   localName;
    private String   namespace;
    private Document ownerDoc;
    private int      childIndex;
    private Node     parent;
    private List     attributes;
    private Map      _namespaces;
    
    protected SimpleNode (Map namespaces) {
        this (null, null, null);
        ownerDoc    = (SimpleDocument) this;
        _namespaces = namespaces;
    }
    
    protected SimpleNode (Document owner, String elementNS, String name) {
        children    = new LinkedList ();
        localName   = name;
        namespace   = elementNS;
        ownerDoc    = owner;
        childIndex  = -2;
        parent      = null;
        attributes  = new LinkedList ();
        _namespaces = new HashMap ();
    }
    
    /*protected SimpleNode (Document owner, Node parentNode, String elementNS, String name) {
        children    = new LinkedList ();
        localName   = name;
        namespace   = elementNS;
        ownerDoc    = owner;
        childIndex  = -2;
        parent      = parentNode;
        attributes  = new LinkedList ();
        _namespaces = new HashMap ();
    }*/
    
    protected SimpleNode () {
        children   = new LinkedList ();
        localName  = null;
        namespace  = null;
        ownerDoc   = (Document) this;
        childIndex = -2;
        parent     = null;
        attributes = new LinkedList ();
    }
    
    public void addNamespace (String prefix, String namespace) {
        _namespaces.put (prefix, namespace);
    }
    
    public Node appendChild (Node newChild) throws DOMException {
        children.add (newChild);
        if (newChild instanceof SimpleNode) {
            incorporateChild ((SimpleNode) newChild, children.size () - 1);
        }
        
        return newChild;
    }
    
    public List cloneChildren (Document newDocument) {
        List     clones    = new LinkedList ();
        Iterator eachChild = children.iterator ();
        Node     child;
        
        while (eachChild.hasNext ()) {
            child = (Node) eachChild.next ();
            if (child instanceof SimpleNode) {
                clones.add (((SimpleNode) child).cloneNode (true, newDocument));
            } else {
                clones.add (child.cloneNode (true));
            }
        }
        
        return clones;
    }
    
    public Node cloneNode (boolean deep) {
        return cloneNode (deep, getOwnerDocument ());
    }
    
    abstract SimpleNode cloneNode (boolean deep, Document newDocument);
    
    public boolean equals (Object other) {
        return isSameNode ((Node) other);
    }
    
    public Map getAllNamespaces () {
        Map all = new HashMap (_namespaces);
        
        if (getParentNode () != null) {
            all.putAll (((SimpleNode) getParentNode ()).getAllNamespaces ());
        }
        
        return all;
    }
    
    public Attr getAttributeNodeNS (String namespaceURI, String localName) throws DOMException {
        return (Attr) getAttributes ().getNamedItemNS (namespaceURI, localName);
    }
    
    public NamedNodeMap getAttributes () {
        return new SimpleNamedNodeMap (attributes);
    }
    
    public int getChildCount () {
        return children.size ();
    }
    
    public NodeList getChildNodes () {
        return new SimpleNodeList (children);
    }
    
    public Node getFirstChild () {
        if (children.size () > 0) {
            return (Node) children.get (0);
        } else {
            return null;
        }
    }

    public Node getLastChild () {
        NodeList children = getChildNodes ();
        
        return children.item (children.getLength () - 1);
    }
    
    public String getLocalName () {
        return localName;
    }
    
    public String getNamespaceForPrefix (String prefix) {
        String namespace = (String) _namespaces.get (prefix);
        
        if (namespace == null && getParentNode () != null) {
            return ((SimpleNode) getParentNode ()).getNamespaceForPrefix (prefix);
        }
        
        return namespace;
    }
    
    public String getNamespaceURI () {
        return namespace;
    }
    
    public Node getNextSibling () {
        if (getParentNode () == null) {
            System.out.println ("Getting next sibling of " + getNodeName () + " but has no parent. childIndex is " + childIndex);
            return null;
        }
        NodeList siblings = getParentNode ().getChildNodes ();
        Node     retVal   = siblings.item (childIndex + 1);
        return retVal;
    }
    
    public String getNodeName () {
        if (namespace != null) {
            return namespace + localName;
        } else {
            return localName;
        }
    }
    
    public Document getOwnerDocument () {
        return ownerDoc;
    }
    
    public Node getParentNode () {
        return parent;
    }
    
    public String getPrefix () {
        return getPrefixForNamespace (namespace);
    }
    
    public String getPrefixForNamespace (String namespace) {
        Iterator eachPrefix = _namespaces.keySet ().iterator ();
        String   prefix;
        
        while (eachPrefix.hasNext ()) {
            prefix = (String) eachPrefix.next ();
            if (_namespaces.get (prefix).equals (namespace)) {
                return prefix;
            }
        }
        if (getParentNode () != null) {
            return ((SimpleNode) getParentNode ()).getPrefixForNamespace (namespace);
        }
        
        return null;
    }
    
    public String getTagName () {
        return getNodeName ();
    }
    
    public boolean hasAttributes () {
        return getAttributes ().getLength () > 0;
    }
    
    public boolean hasChildNodes () {
        return getChildCount () > 0;
    }
    
    public SimpleNode importInto (Document newDocument) {
        return (SimpleNode) cloneNode (true, newDocument);
    }
    
    protected void incorporateChild (SimpleNode newChild, int index) {
        newChild.setChildIndex (index);
        newChild.setParent     (this);
        
        if (newChild.getOwnerDocument () != getOwnerDocument ()) {
            int      preNum   = 1;
            Iterator eachNS;
            String   space;
            String   prefix;
            
            eachNS = newChild.getAllNamespaces ().values ().iterator ();
            while (eachNS.hasNext ()) {
                space = (String) eachNS.next ();
                if (getPrefixForNamespace (space) == null) {
                    prefix = newChild.getPrefixForNamespace (space);
                    while (getNamespaceForPrefix (prefix) != null) {
                        prefix = "ns" + preNum;
                        preNum += 1;
                    }
                    addNamespace (prefix, space);
                }
            }
        }
    }
    
    private static boolean equalOrNull (Object first, Object second) {
        return (first == null && second == null) || (first != null && second != null && first.equals (second));
    }
    
    public boolean isEqualNode (Node other) {
/*System.out.println ("Compare type " + other.getNodeType () + " with " + getNodeType ());
System.out.println ("Compare namespace " + other.getNamespaceURI () + " with " + getNamespaceURI ());
System.out.println ("Compare name " + other.getLocalName () + " with " + getLocalName ());
System.out.println ("Compare value " + other.getNodeValue () + " with " + getNodeValue ());
System.out.println ("Compare children " + other.getChildNodes ().getLength () + " with " + getChildNodes ().getLength ());*/
        if (!(other.getNodeType () == getNodeType ()
        && equalOrNull (getNamespaceURI (), other.getNamespaceURI ())
        && equalOrNull (getLocalName (), other.getLocalName ())
        && equalOrNull (getNodeValue (), other.getNodeValue ())
        && getChildNodes ().getLength () == other.getChildNodes ().getLength ())) {
            return false;
        }
        NodeList children = getChildNodes ();
        for (int index = 0; index < children.getLength (); index += 1) {
            if (!children.item (index).isEqualNode (other.getChildNodes ().item (index))) {
                return false;
            }
        }
        
        return true;
    }
    
    public boolean isSameNode (Node otherNode) {
        SimpleNode other;
        
        if (otherNode == this) {
            return true;
        }
        if (otherNode instanceof SimpleNode) {
            other = (SimpleNode) otherNode;
            if (other.ownerDoc == ownerDoc) {
                if ((other.parent == null && parent == null) || (other.parent != null && parent != null && other.parent.isSameNode (parent))) {
                    if (equalOrNull (other.localName, localName) && equalOrNull (other.namespace, namespace)) {
                        if (other.childIndex == childIndex) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }
    
    public void setAttributeNS (String namespaceURI, String qualifiedName, String value) throws DOMException {
        setAttributeNodeNS (new SimpleAttr (getOwnerDocument (), namespaceURI, qualifiedName, value));
    }
    
    public Attr setAttributeNodeNS (Attr newAttr) throws DOMException {
        attributes.add (newAttr);
        return newAttr;
    }
    
    void setChildIndex (int newIndex) {
        childIndex = newIndex;
    }
    
    void setContent (List newElements) {
        children = newElements;
    }
    
    void setLocalName (String newName) {
        localName = newName;
    }
    
    void setParent (Node newParent) {
        parent = newParent;
    }
    
    public String toString () {
        return getNodeName ();
    }
}
