package org.pasoa.query.provenance;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import org.pasoa.common.Constants;
import org.pasoa.common.Retrieval;
import org.pasoa.common.structure.GlobalPAssertionKey;
import org.pasoa.common.structure.InteractionKey;
import org.pasoa.common.structure.Link;
import org.pasoa.common.structure.ObjectID;
import org.pasoa.common.structure.PAssertionDataKey;
import org.pasoa.common.structure.deserialiser.PStructureDeserialiser;
import org.pasoa.dommap.DOMDeserialiser;
import org.pasoa.query.dom.QStructureDocument;
import org.w3c.dom.Element;

public class ProvenanceQueryDeserialiser {
    public ProvenanceQueryDeserialiser () {
    }
    
    public ProvenanceQuery deserialiseProvenanceQuery (Element serialised, Retrieval retrieval, String thisStoreURL, QStructureDocument data) throws Exception {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser           parse  = new DOMDeserialiser (Constants.PQUERY_NS);
        DocumentSetAccessorSearch qdh    = deserialiseQueryDataHandle (parse.single (serialised, "queryDataHandle"), retrieval, thisStoreURL, data);
        DocumentCheck             filter = deserialiseRelationshipTargetFilter (parse.single (serialised, "relationshipTargetFilter"));
        
        return new ProvenanceQuery (retrieval, thisStoreURL, qdh, filter);
    }
    
    public ProvenanceQueryResults deserialiseProvenanceQueryResults (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser parse             = new DOMDeserialiser (Constants.PQUERY_NS);
        Element         start             = parse.single (serialised, "start");
        Iterator        eachStart         = parse.iterator (start, "pAssertionDataKey");
        Iterator        eachFull          = parse.iterator (serialised, "fullRelationship");
        Collection      starts            = new LinkedList ();
        Collection      fullRelationships = new LinkedList ();
        
        while (eachStart.hasNext ()) {
            starts.add (deserialisePAssertionDataKey ((Element) eachStart.next ()));
        }
        while (eachFull.hasNext ()) {
            fullRelationships.add (deserialiseFullRelationship ((Element) eachFull.next ()));
        }

        return new ProvenanceQueryResults (starts, fullRelationships);
    }

    public FullRelationship deserialiseFullRelationship (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser parse    = new DOMDeserialiser (Constants.PQUERY_NS);
        ObjectID        subject  = deserialiseObjectID (parse.single (serialised, "fullSubjectId"));
        String          relation = parse.text (serialised, "relation");
        String          lpaid    = parse.text   (serialised, "localPAssertionId");
        ObjectID        object   = deserialiseObjectID (parse.single (serialised, "fullObjectId"));
        
        return new FullRelationship (subject, relation, lpaid, object);
    }
    
    public PAssertionDataKey deserialisePAssertionDataKey (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser        parse       = new DOMDeserialiser (Constants.PSTRUCT_NS);
        PStructureDeserialiser pstruct     = new PStructureDeserialiser ();
        InteractionKey         interaction = pstruct.getInteractionRecordID (parse.single (serialised, "interactionKey"));
        String                 viewKind    = parse.text   (serialised, "viewKind");
        String                 lpaid       = parse.text   (serialised, "localPAssertionId");
        Element                accessor    = parse.single (serialised, "dataAccessor");
        
        return new PAssertionDataKey (new GlobalPAssertionKey (interaction, viewKind, lpaid), accessor);
    }
    
    public ObjectID deserialiseObjectID (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        PAssertionDataKey padk  = deserialisePAssertionDataKey (serialised);
        DOMDeserialiser   parse = new DOMDeserialiser (Constants.PSTRUCT_NS);
        String            param = parse.text (serialised, "parameterName");
        String            store = parse.text (parse.single (serialised, "objectLink"), "provenanceStoreRef");
        
        return new ObjectID (padk, param, new Link (store));
    }
    
    public DocumentSetAccessorSearch deserialiseQueryDataHandle (Element serialised, Retrieval retrieval, String thisStoreURL, QStructureDocument data) throws Exception {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser     parse      = new DOMDeserialiser (Constants.PQUERY_NS);
        Element             search     = parse.single (parse.single (serialised, "search"), "xpathSearch");
        String              reference  = parse.text (search, "path");
        Map                 namespaces = deserialiseNamespaceMappings (search);
        PStructureReference searchOver = deserialisePStructureReference (parse.single (serialised, "pStructureReference"), retrieval, thisStoreURL, data);
        
        return new DocumentSetAccessorSearch (reference, namespaces, "", searchOver);
    }
    
    public DocumentCheck deserialiseRelationshipTargetFilter (Element serialised) {
        if (serialised == null) {
            return null;
        }
        DOMDeserialiser parse      = new DOMDeserialiser (Constants.PQUERY_NS);
        Element         search     = parse.single (parse.single (serialised, "check"), "xpathSearch");
        String          reference  = parse.text (search, "path");
        Map             namespaces = deserialiseNamespaceMappings (search);
        
        return new DocumentCheck (reference, namespaces, "");
    }
    
    public PStructureReference deserialisePStructureReference (Element serialised, Retrieval retrieval, String thisStoreURL, QStructureDocument data) throws Exception {
        if (serialised == null) {
            return null;
        }
        
        DOMDeserialiser parse    = new DOMDeserialiser (Constants.PQUERY_NS);
        boolean         contents = parse.single (serialised, "storeContents") != null;
        ProvenanceQuery pQuery   = deserialiseProvenanceQuery (parse.single (serialised, "provenanceQuery"), retrieval, thisStoreURL, data);
        //TransitiveClosure
        
        if (contents) {
            return new ProvenanceStoreContents (data);
        }
        if (pQuery != null) {
            return pQuery;
        }
        return null;
    }
    
    public Map deserialiseNamespaceMappings (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        Map             mappings = new HashMap ();
        DOMDeserialiser parse    = new DOMDeserialiser (Constants.PQUERY_NS);
        Iterator        eachMap  = parse.iterator (serialised, "namespaceMapping");
        
        while (eachMap.hasNext ()) {
            deserialiseNamespaceMapping ((Element) eachMap.next (), mappings);
        }
        
        return mappings;
    }
    
    public void deserialiseNamespaceMapping (Element serialised, Map addTo) {
        if (serialised == null) {
            return;
        }
        
        DOMDeserialiser parse     = new DOMDeserialiser (Constants.PQUERY_NS);
        String          prefix    = parse.text (serialised, "prefix");
        String          namespace = parse.text (serialised, "namespace");
        
        addTo.put (prefix, namespace);
    }
}
