/*Copyright (c) 2005 University of Southampton
 
 Permission is hereby granted, free of charge, 
 to any person obtaining a copy of this 
 software and associated documentation files 
 (the "Software"), to deal in the Software without 
 restriction, including without limitation the rights 
 to use, copy, modify, merge, publish, distribute, 
 sublicense, and/or sell copies of the Software, and
 to permit persons to whom the Software is furnished 
 to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice 
 shall be included in all copies or substantial 
 portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY 
 OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO 
 EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
 FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
 AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE 
 OR OTHER DEALINGS IN THE SOFTWARE.
 -------------------------------------------------------*/
package org.pasoa.test.examples;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URL;
import java.util.LinkedList;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.pasoa.common.Constants;
import org.pasoa.common.structure.Record;
import org.pasoa.common.structure.ActorStatePAssertion;
import org.pasoa.common.structure.GlobalPAssertionKey;
import org.pasoa.common.structure.InteractionKey;
import org.pasoa.common.structure.InteractionPAssertion;
import org.pasoa.common.structure.ObjectID;
import org.pasoa.common.structure.Relationship;
import org.pasoa.dommap.DOMWriter;
import org.pasoa.util.ClientLib;
import org.pasoa.util.httpsoap.WSAddressEndpoint;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * Demonstrates how to use the ClientLib to record interaction p-assertions,
 * actor state p-assertions, relationship p-assertions and submission finished
 * messages using the ClientLib.
 * 
 * @author pgroth
 * 
 */
public class ClientLibExample
{
    public static void main(String[] args) throws Exception
    {
        if (args.length != 1) {
            System.out.println ("Usage: java org.pasoa.test.examples.ClientLibExample <provenance store url>");
            System.out.println ("       example url: http://localhost:8080/preserv");
            System.out.println ("       There should be NO trailing / behind preserv");
            System.exit(0);
        }
        
        // Here we create the urls for recording and querying the provenance store
        // send record messages to the record port (provenanceStore)
        // send query messages to the xquery port (proveanceStoreQuery)
        URL provenanceStore = new URL (args[0] + "/record");
        URL provenanceStoreQuery = new URL (args[0] + "/xquery");
        
        ClientLibExample cle = new ClientLibExample();
        
        System.out.println("ClientLibExample");
        System.out.println("The scenario:");
        System.out.println("      We simulate a sender sending a message");
        System.out.println("      to a receiver. The sender uses part of");
        System.out.println("      its state in the the message so it ");
        System.out.println("      makes a relationship p-assertion between");
        System.out.println("      its actor state p-assertion and its ");
        System.out.println("      interaction p-assertion");
        System.out.println();
        System.out.println();
        System.out.println("Press Enter to continue...");
        BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
        br.readLine();

        
        System.out.println("We now create the interaction key");
        
        // Create addresses for the source and sink of the 
        // interaction.
        WSAddressEndpoint source = new WSAddressEndpoint ("http://www.sender.com/sender");
        WSAddressEndpoint sink   = new WSAddressEndpoint ("http://www.receiver.com/receiver");
        
        // Create an interactionId, this should be unique!
        String interactionId = "uniqueId" + System.currentTimeMillis();
        InteractionKey ik = new InteractionKey (source.getElement(), sink.getElement(), interactionId);
        
        System.out.println ("The interaction key created is: ");
        System.out.println ("   " + ik.toString());
        System.out.println ();
        
        System.out.println ("We are the sender/client view of the interaction");
        String vk = Constants.CLIENT_VIEW;
        System.out.println ();
        
        System.out.println ("The asserter is identified by the following url: http://www.users.com/SomeGuy");
        WSAddressEndpoint asserter = new WSAddressEndpoint ("http://www.users.com/SomeGuy");
        System.out.println ("Press Enter to continue...");
        br.readLine();
        
        
        System.out.println("Building p-assertions...");
        
        InteractionPAssertion ipa = cle.createInteractionPAssertion();
        ActorStatePAssertion  asa = cle.createActorStatePAssertion();
        Relationship          rel = cle.createRelationshipPAssertion (ik, vk); 
        
        System.out.println("Three p-assertions have been created");
        System.out.println("Creating Record objects for each p-assertion");
        
        Record recIpa = new Record (ipa, ik, vk, asserter.getElement());
        Record recAsa = new Record (asa, ik, vk, asserter.getElement());
        Record recRel = new Record (rel, ik, vk, asserter.getElement());
        
        
        ClientLib clientLib = new ClientLib();

        System.out.println("Recording the p-assertions in provenance store " + provenanceStore);
        
        clientLib.record(recIpa, provenanceStore);
        clientLib.record(recAsa, provenanceStore);
        clientLib.record(recRel, provenanceStore);
        
        
        System.out.println ("Three p-assertions recorded");
        System.out.println ("Press Enter to continue...");
        br.readLine();
        System.out.println ();
        System.out.println ("Querying the content of the provenance store");
        System.out.println ("    The entire content of the store will be displayed");
        System.out.println ("Press any key to continue...");
        br.readLine();
        String xquery = "<result>{for $n in $ps:pstruct return $n}</result>";
        Document result = clientLib.query(xquery, provenanceStoreQuery);
        System.out.println (DOMWriter.writeToString (result));
        System.out.println ();
        System.out.println ("Done");
        System.out.println ();
        System.out.println ();
        System.out.println ();
        System.out.println ("Thanks for using ClientLibExample!");
        System.out.println ("Have fun creating your own provenance aware application!");
        System.out.println ();
    }
    
    private InteractionPAssertion createInteractionPAssertion() 
        throws Exception
    {
        // Create an interaction p-assertion
        // First we make a local p-assertion id and then
        // we make a documentationStyle. In this case we 
        // call it verbatium.
        //
        // In most cases, you'll be grabing the messageBody from the message
        // being sent between parties. So a SOAP message, or a CORBA message.
        // With this example we'll just use a hard coded message body.
        
        String localPAssertionId = "1";
        String documentationStyle = "http://www.pasoa.org/schemas/ontologies/DocumentationStyle.Verbatim";
        
        // The contents of a message
        String message = "<hello>Blah</hello>";
        
        // Convert it into a DOM Element
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        Document msgDoc = db.parse(new InputSource(new StringReader(message)));
        Element messageBody = msgDoc.getDocumentElement();
        
        InteractionPAssertion ipa = new InteractionPAssertion(localPAssertionId, documentationStyle, messageBody);
        
        return ipa;
    }
    
    private ActorStatePAssertion createActorStatePAssertion() 
        throws Exception
    {
        // Create an actor state p-assertion
        // Just like the interaction p-assertion this p-assertion
        // needs a local p-assertion id. Remember, all the p-assertions
        // in one view need a different id. Therefore, we give this assertion
        // the id of "2" instead of "1".
        //
        // Again you'll typically be getting some state from the actor,
        // translating it to XML to create the actor state p-assertion
        // In this example, we just use a hard coded string.
        
        String localPAssertionId = "2";
        
        // The contents of the assertion
        String actorStateBody = "<state>Blah</state>";
        
        // Convert it into a DOM Element
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        Document msgDoc = db.parse(new InputSource(new StringReader(actorStateBody)));
        Element stateBody = msgDoc.getDocumentElement();
        
        ActorStatePAssertion asa = new ActorStatePAssertion(localPAssertionId, stateBody);
        
        return asa;
    }
    
    private Relationship createRelationshipPAssertion (InteractionKey ik, String view) 
         throws Exception
    {
        // Create a relationship p-assertion
        // Again a different local p-assertion id
        //
        // We'll create a "usage" relationship between the interaction p-assertion
        // and the actor state p-assertion. This relationship says that 
        // message represented by interaction p-assertion "1" used the actor state
        // represented by actor state p-assertion "2".
        // There are no data accessors or links so we pass null.
     
        // Create the information to identify the subject of the relationship
        // Remember, parameter names must be identified and they need to be URIs
        String localPAssertionId = "3";
        String subjectLocalPAssertionId = "1";
        String subjectParameterName = "http://www.pasoa.org/schemas/ontologies/relationships/usage.user";
        
        // Create the information to identify the object of the relationship
   
        String objectLocalPAssertionId = "2"; // points to the actor state p-assertion
        
        // We get the interaction key and view where the actor state p-assertion 
        // is from the parameters to this method. In this case, the actor state p-assertion
        // is in the same interaction record and view as the interaction p-assertion. 
        // In many other cases, the object of the relation could be in another interaction record
        // and in another view.
        GlobalPAssertionKey gpak = new GlobalPAssertionKey(ik, view, objectLocalPAssertionId);
        String objectParameterName = "http://www.pasoa.org/schemas/ontologies/relationships/usage.usedItem";
        ObjectID objId = new ObjectID (gpak, objectParameterName, null, null);
        
        // We add the objId to the list of objects. We only have one objectId here
        // but when making another type of relationship more than one objectId may
        // be required
        LinkedList objectIds = new LinkedList();
        objectIds.add (objId);
        
        // Create the "use" relation. Again this should be a URI
        String relation = "http://www.pasoa.org/schemas/ontologies/relationships/use";         

        // Finally, create the relationship object and return it.
        Relationship rel = new Relationship(localPAssertionId, 
                                            subjectLocalPAssertionId, 
                                            null, 
                                            subjectParameterName, 
                                            relation, 
                                            objectIds);
        
        return rel;

    }

}
