/********************************************************************************
 * Created on Sep 6, 2005
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.test.examples.average;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import org.pasoa.common.Constants;
import org.pasoa.common.structure.Record;
import org.pasoa.test.RemoteTestbed;
import org.pasoa.test.Testbed;
import org.pasoa.util.httpsoap.WSAddressEndpoint;
import org.w3c.dom.Element;

/**
 * Represents the process documentation of a simple workflow in which four
 * actors (gui, averager, divider and store) interact.  Using this class you
 * can record all p-assertions about this workflow into a remote or local provenance
 * store, by giving the appropriate Testbed instance.
 */
public class AverageExample {
    // Example namespace
    public static final String ns = "http://www.example.com";
    
    // Actors
    public static final String gui      = ns + "/actors/gui";
    public static final String averager = ns + "/actors/averager";
    public static final String divider  = ns + "/actors/divider";
    public static final String store    = ns + "/actors/store";
    
    // Tracers
    public static final String tracer1 = "process://1";
    public static final String tracer2 = "process://sub/1";
    
    // Message contents - location where data stored
    public static final String locate  = "/home/sm/data/file1";
    // Message contents - namespace of example message elements
    public static final String xmlns   = "xmlns:ex = \"" + ns + "\"";
    // Message contents - p-header containing tracer added by GUI
    public static final String header1 = "<ex:envelope " + xmlns + "><ph:pheader xmlns:ph = \"" + Constants.PHEADER_NS + "\"><ph:interactionMetaData><ph:tracer>" + tracer1 + "</ph:tracer></ph:interactionMetaData></ph:pheader>";
    // Message contents - p-header containing tracer added by Averager
    public static final String header2 = "<ex:envelope " + xmlns + "><ph:pheader xmlns:ph = \"" + Constants.PHEADER_NS + "\"><ph:interactionMetaData><ph:tracer>" + tracer1 + "</ph:tracer><ph:tracer>" + tracer2 + "</ph:tracer></ph:interactionMetaData></ph:pheader>";
    public static final String footer  = "</ex:envelope>";
    // Message contents - message between GUI and Averager
    public static final String msg1    = header1 + "<ex:average><ex:first   value = \"5\" /><ex:second   ex:value = \"7\"/></ex:average>" + footer;
    // Message contents - message between Averager and Divider
    public static final String msg2    = header2 + "<ex:divide><ex:divisor value = \"12\"/><ex:dividend ex:value = \"2\"/></ex:divide>" + footer;
    // Message contents - message between Divider and Averager
    public static final String msg3    = header2 + "<ex:answer value = \"6\"/>" + footer;
    // Message contents - message between Averager and GUI
    public static final String msg4    = header1 + "<ex:answer value = \"6\"/>" + footer;
    // Message contents - message between GUI and Store
    public static final String msg5    = header1 + "<ex:store><ex:location>" + locate + "</ex:location><ex:data>6</ex:data></ex:store>" + footer;
    
    // Ontologies
    public static final String relOntology = "http://www.ecs.soton.ac.uk/~sm/PASOA/relations.n3#";
    public static final String tmpOntology = "http://twiki.pasoa.ecs.soton.ac.uk/pub/PASOA/Ontologies/temporal.owl#";
    public static final String cslOntology = "http://twiki.pasoa.ecs.soton.ac.uk/pub/PASOA/Ontologies/causal.owl#";
    
    // Relations
    public static final String sumOf       = relOntology + "sumOf";
    public static final String divisionOf  = relOntology + "divisionOf";
    public static final String copyOf      = relOntology + "copyOf";
    public static final String averageOf   = relOntology + "averageOf";
    
    // Parameters (sumOf)
    public static final String part        = ns + "/parameters/part";
    public static final String sum         = ns + "/parameters/sum";
    // Parameters (divisionOf)
    public static final String divisor     = ns + "/parameters/divisor";
    public static final String dividend    = ns + "/parameters/dividend";
    public static final String quotient    = ns + "/parameters/quotient";
    // Parameters (copyOf)
    public static final String compared    = ns + "/parameters/compared";
    // Parameters (averageOf)
    public static final String item        = ns + "/parameters/item";
    public static final String average     = ns + "/parameters/average";
    
    // Data Accessors
    public static final String firstPath    = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:average[0]/ex:first[0]";
    public static final String secondPath   = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:average[0]/ex:second[0]";
    public static final String divisorPath  = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:divide[0]/ex:divisor[0]";
    public static final String dividendPath = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:divide[0]/ex:dividend[0]";
    public static final String quotientPath = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:answer[0]";
    public static final String averagePath  = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:answer[0]";
    public static final String storePath    = "/ps:interactionPAssertion[0]/ex:envelope[0]/ex:store[0]/ex:data[0]";
    
    public Map spaces;
    
    public AverageExample () {
        spaces = new HashMap ();
        
        spaces.put ("ps",  Constants.PSTRUCT_NS);
        spaces.put ("pq",  Constants.PQUERY_NS);
        spaces.put ("ph",  Constants.PHEADER_NS);
        spaces.put ("wsa", WSAddressEndpoint.wsaNamespace);
        spaces.put ("ex",  ns);
    }
    
    /**
     * Record all the p-assertions about the workflow into the store at the 
     * URL (record port) given by the first argument.
     */
    public static void main (String[] arguments) throws Exception {
        new AverageExample ().record (new URL (arguments [0]));
    }

    /**
     * Record all the p-assertions about the workflow into the store at the 
     * given URL (record port).
     */
    public void record (URL storeRecordURL) throws Exception {
        record (new RemoteTestbed (storeRecordURL, null, null));
    }
    
    /**
     * Record all the p-assertions about the workflow into the store used by
     * the given testbed.
     */
    public void record (Testbed testbed) throws Exception {
        Element firstArg    = testbed.createDataAccessor (firstPath,    "ex", ns);
        Element secondArg   = testbed.createDataAccessor (secondPath,   "ex", ns);
        Element divisorArg  = testbed.createDataAccessor (divisorPath,  "ex", ns);
        Element dividendArg = testbed.createDataAccessor (dividendPath, "ex", ns);
        Element quotientArg = testbed.createDataAccessor (quotientPath, "ex", ns);
        Element averageArg  = testbed.createDataAccessor (averagePath,  "ex", ns);
        Element storedArg   = testbed.createDataAccessor (storePath,    "ex", ns);
        Record  result;
        Record  inter1, inter2, inter3, inter4, inter5;
        
        inter1  = testbed.recordInteraction (gui,      averager, msg1);
        inter2  = testbed.recordInteraction (averager, divider,  msg2);
        inter3  = testbed.recordInteraction (divider,  averager, msg3);
        inter4  = testbed.recordInteraction (averager, gui,      msg4);
        inter5  = testbed.recordInteraction (gui,      store,    msg5);
        testbed.recordRelationship (inter2, divisorArg,  sum,      inter1, firstArg,    part,     inter1, secondArg,   part,     sumOf);
        testbed.recordRelationship (inter3, quotientArg, quotient, inter2, divisorArg,  divisor,  inter2, dividendArg, dividend, divisionOf);
        testbed.recordRelationship (inter4, averageArg,  compared, inter3, quotientArg, compared, copyOf);
        testbed.recordRelationship (inter4, averageArg,  average,  inter1, firstArg,    item,     inter1, secondArg,   item,     averageOf);
        testbed.recordRelationship (inter5, storedArg,   compared, inter4, averageArg,  compared, copyOf);
        
        // Wait for the provenance store to process everything
        Thread.sleep (5000);
    }
}
