/*Copyright (c) 2005 University of Southampton
         
Permission is hereby granted, free of charge, 
to any person obtaining a copy of this 
software and associated documentation files 
(the "Software"), to deal in the Software without 
restriction, including without limitation the rights 
to use, copy, modify, merge, publish, distribute, 
sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished 
to do so, subject to the following conditions:
            
The above copyright notice and this permission notice 
shall be included in all copies or substantial 
portions of the Software.
         
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY 
OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO 
EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN 
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE 
OR OTHER DEALINGS IN THE SOFTWARE.
-------------------------------------------------------*/

package org.pasoa.util;


import java.net.URL;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.pasoa.common.structure.ActorStatePAssertion;
import org.pasoa.common.structure.InteractionPAssertion;
import org.pasoa.common.structure.InteractionKey;
import org.pasoa.common.structure.ObjectID;
import org.pasoa.common.structure.Record;
import org.pasoa.common.structure.RecordSf;
import org.pasoa.common.structure.Relationship;
import org.pasoa.dommap.DOMWriter;
import org.pasoa.common.Constants;
import org.pasoa.util.httpsoap.DefaultDocumentClient;
import org.pasoa.util.httpsoap.ProvenanceQueryRequestDocument;

import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * ClientLib provides a number of convenience methods for recording,
 * querying, (and hopefully in the future) managing provenance stores. 
 * The library takes care of serialization/deserialization from Java objects
 * to SOAP.
 * <p>
 * The two methods of most use are probably <code>record (Record rec, URL provenanceStore)</code>
 * which allows you to build a Record object and send that to a provenance
 * store and <code>query (String xquery, URL provenanceStore)</code> which allows you to query
 * a particular provenance store with an xquery string. 
 * <p>
 * The documentation assumes you are familiar with the Provenance Modelling and
 * Functionality section of the <a href="http://eprints.ecs.soton.ac.uk/11310/">Architecture for Provenance Systems</a> document.
 * 
 * @author pgroth
 *
 */
public class ClientLib {
    
    
    /**
     * Record a Record object in the specified provenance store.
     * 
     * @param rec the Record object
     * @param provenanceStore the URL of the provenance store
     * @return a DOM Document containing the response from the provenance store
     * @throws Exception
     */
    public Document record (Record rec, URL provenanceStore) throws Exception {
        String request = makeRecordDocument (rec);
        //System.out.println ("Request: " + request);
        DefaultDocumentClient ddc = new DefaultDocumentClient ();
        Document response = ddc.invoke (provenanceStore, request);
        return response;
    }
    
    /**
     * Record an already generated Record SOAP message in the specified provenance store
     * This SOAP message must conform to PRecord.wsdl
     * 
     * @param soap the record soap message to send to the provenance store
     * @param provenanceStore the URL of the provenance store
     * @return a DOM Document containg the response from the provenance store
     * @throws Exception
     */
    public Document record (String soap, URL provenanceStore) throws Exception {
        DefaultDocumentClient ddc = new DefaultDocumentClient ();
        Document response = ddc.invoke (provenanceStore, soap);
        return response;
    }
    
    /**
     * Generates a Record SOAP message from the given parameters.
     * 
     * @param irid the key that identifies the interaction
     * @param viewKind the ViewKind as specified by Constants.CLIENT_VIEW or Constants.SERVICE_VIEW
     * @param asserter the asserter as a DOM element
     * @param content the content (i.e. interactionPAssertion, actorStatePassertion, relationshipPAssertion
     *                  or submissionFinished) already seralized to XML
     *                  
     * @return a Record SOAP message as a string
     */
    public String makeRecordDocument (InteractionKey irid, String viewKind, Element asserter, String content) {
        StringBuffer sb = new StringBuffer ();
        sb.append ("<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">");
        
        sb.append ("<soapenv:Header>");
        sb.append ("</soapenv:Header>");
        sb.append ("<soapenv:Body>");
        
        sb.append ("<pr:record xmlns:pr=\"" + Constants.PRECORD_NS + "\" xmlns:ps=\"" + Constants.PSTRUCT_NS + "\" xmlns:psx=\"" + Constants.WPSTRUCT_NS + "\" xmlns:ph=\"" + Constants.PHEADER_NS + "\">");
        sb.append ("<pr:identifiedContent>");
        sb.append ("<ps:interactionKey>");
        sb.append ("<ps:messageSource>");
        sb.append (DOMWriter.writeToString(irid.getMessageSource()));
        sb.append ("</ps:messageSource>");
        sb.append ("<ps:messageSink>");
        sb.append (DOMWriter.writeToString(irid.getMessageSink()));
        sb.append ("</ps:messageSink>");
        sb.append ("<ps:interactionId>");
        sb.append (irid.getInteractionID ());
        sb.append ("</ps:interactionId>");
        sb.append ("</ps:interactionKey>");
        
        sb.append ("<ps:viewKind>");
        sb.append (viewKind);
        sb.append ("</ps:viewKind>");
        sb.append ("<ps:asserter>");
        sb.append (DOMWriter.writeToString(asserter));  
        sb.append ("</ps:asserter>");
        sb.append ("<pr:content>");
        sb.append (content);
        sb.append ("</pr:content>");
        sb.append ("</pr:identifiedContent>");
        sb.append ("</pr:record>");
        sb.append ("</soapenv:Body>");
        sb.append ("</soapenv:Envelope>");
        return sb.toString ();
    }
    
    /**
     * Given a Record object generate a Record SOAP message
     * 
     * @param rec a Record object
     * @return a Record SOAP message as a String
     */
    public String makeRecordDocument (Record rec) {
        
        StringBuffer sb = new StringBuffer ();
        sb.append ("<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">");
        
        sb.append ("<soapenv:Header>");
        sb.append ("</soapenv:Header>");
        sb.append ("<soapenv:Body>");
        
        sb.append ("<pr:record xmlns:pr=\"" + Constants.PRECORD_NS + "\" xmlns:ps=\"" + Constants.PSTRUCT_NS + "\" xmlns:psx=\"" + Constants.WPSTRUCT_NS + "\" xmlns:ph=\"" + Constants.PHEADER_NS + "\">");
        sb.append ("<pr:identifiedContent>");
        sb.append ("<ps:interactionKey>");
        sb.append ("<ps:messageSource>");
        sb.append (DOMWriter.writeToString(rec.getIdentifiers().getMessageSource()));
        sb.append ("</ps:messageSource>");
        sb.append ("<ps:messageSink>");
        sb.append (DOMWriter.writeToString(rec.getIdentifiers().getMessageSink()));
        sb.append ("</ps:messageSink>");
        sb.append ("<ps:interactionId>");
        sb.append (rec.getIdentifiers().getInteractionID ());
        sb.append ("</ps:interactionId>");
        sb.append ("</ps:interactionKey>");
        
        sb.append ("<ps:viewKind>");
        sb.append (rec.getViewKind());
        sb.append ("</ps:viewKind>");
        sb.append ("<ps:asserter>");
        sb.append (DOMWriter.writeToString(rec.getAsserter()));  
        sb.append ("</ps:asserter>");
        sb.append ("<pr:content>");
        
        if (rec.getPAssertion () instanceof InteractionPAssertion) {
            InteractionPAssertion ipa = (InteractionPAssertion) rec.getPAssertion ();
            sb.append ("<ps:interactionPAssertion>");
            sb.append ("<ps:localPAssertionId>");
            sb.append (ipa.getLocalPAssertionID ());
            sb.append ("</ps:localPAssertionId>");
            sb.append ("<ps:documentationStyle>");
            sb.append (ipa.getDocumentationStyle ());
            sb.append ("</ps:documentationStyle>");
            sb.append ("<ps:content>");
            String messageBody = DOMWriter.writeToString (ipa.getMessageBody ());
            sb.append (messageBody);
            sb.append ("</ps:content>");
            sb.append ("</ps:interactionPAssertion>");
        } else if (rec.getPAssertion () instanceof ActorStatePAssertion) {
            ActorStatePAssertion apa = (ActorStatePAssertion) rec.getPAssertion ();
            sb.append ("<ps:actorStatePAssertion>");
            sb.append ("<ps:localPAssertionId>");
            sb.append (apa.getLocalPAssertionID ());
            sb.append ("</ps:localPAssertionId>");
            sb.append ("<ps:content>");
            //DOMImplementation impl = apa.getActorStateBody().getOwnerDocument().getImplementation();
            //DOMImplementationLS implLS =  (DOMImplementationLS) impl.getFeature("LS","3.0");
            //LSSerializer writer=  implLS.createLSSerializer();
            //writer.getDomConfig().setParameter("xml-declaration", new Boolean(false));
            String messageBody = DOMWriter.writeToString (apa.getActorStateBody ());
            sb.append (messageBody);
            sb.append ("</ps:content>");
            sb.append ("</ps:actorStatePAssertion>");
        } else if (rec.getPAssertion () instanceof Relationship) {
            Relationship rel = (Relationship) rec.getPAssertion ();
            sb.append ("<ps:relationshipPAssertion>");
            sb.append ("<ps:localPAssertionId>");
            sb.append (rel.getLocalPAssertionID ());
            sb.append ("</ps:localPAssertionId>");
            sb.append ("<ps:subjectId>");
            sb.append ("<ps:localPAssertionId>");
            sb.append (rel.getSubject().getLocalPAssertionID());
            sb.append ("</ps:localPAssertionId>");
            
            if (rel.getSubject().getDataAccessor() != null) {
                sb.append ("<ps:dataAccessor>");
                sb.append (DOMWriter.writeToString (rel.getSubject().getDataAccessor()));
                sb.append ("</ps:dataAccessor>");
            }
            
            sb.append ("<ps:parameterName>");
            sb.append (rel.getSubject().getParameterName());
            sb.append ("</ps:parameterName>");
            sb.append ("</ps:subjectId>");
            
            sb.append ("<ps:relation>");
            sb.append (rel.getRelation ());
            sb.append ("</ps:relation>");
            
            List objectIDs = rel.getObjectIDs ();
            Iterator iter = objectIDs.iterator ();
            
            while (iter.hasNext ()) {
                ObjectID objID = (ObjectID)iter.next ();
                sb.append ("<ps:objectId>");
                
                sb.append ("<ps:messageSource>");
                sb.append (DOMWriter.writeToString (objID.getGlobalPAssertionKey ().getInteractionKey ().getMessageSource ()));
                sb.append ("</ps:messageSource>");
                sb.append ("<ps:messageSink>");
                sb.append (DOMWriter.writeToString (objID.getGlobalPAssertionKey ().getInteractionKey ().getMessageSink ()));
                sb.append ("</ps:messageSink>");
                sb.append ("<ps:interactionId>");
                sb.append (objID.getGlobalPAssertionKey ().getInteractionKey ().getInteractionID ());
                sb.append ("</ps:interactionId>");
                sb.append ("<ps:viewKind>");
                sb.append (objID.getGlobalPAssertionKey ().getViewName ());
                sb.append ("</ps:viewKind>");
                sb.append ("<ps:localPAssertionId>");
                sb.append (objID.getGlobalPAssertionKey ().getLocalPAssertionID ());
                sb.append ("</ps:localPAssertionId>");
                
                if (objID.getObjectDataAccessor() != null) {
                    sb.append ("<ps:dataAccessor>");
                    sb.append (DOMWriter.writeToString (objID.getObjectDataAccessor ()));
                    sb.append ("</ps:dataAccessor>");
                }
                
                sb.append ("<ps:parameterName>");
                sb.append (objID.getParameterName ());
                sb.append ("</ps:parameterName>");
                
                if (objID.getLink() != null) {
                    sb.append ("<ps:objectLink>");
                    sb.append ("<ps:provenanceStoreRef>");
                    sb.append (objID.getLink().getStoreURL());
                    sb.append ("</ps:provenanceStoreRef>");
                    sb.append ("</ps:objectLink");
                }
                
                sb.append ("</ps:objectId>");
            }
            sb.append ("</ps:relationshipPAssertion>");
        }
        
        if (rec instanceof RecordSf) {
            RecordSf rsf = (RecordSf)rec;
            sb.append ("<pr:submissionFinished>");
            sb.append (rsf.getNumberOfExpectedPAssertions ());
            sb.append ("</pr:submissionFinished>");
        }
        
        
        
        sb.append ("</pr:content>");
        sb.append ("</pr:identifiedContent>");
        sb.append ("</pr:record>");
        sb.append ("</soapenv:Body>");
        sb.append ("</soapenv:Envelope>");
        return sb.toString ();
    }
    
    /**
     * Send an XQuery to the specified provenance store.  See
     * http://twiki.pasoa.ecs.soton.ac.uk/bin/view/PASOA/QueryDocumentation
     * for full documentation on how to use this interface.
     * 
     * @param xquery the xquery string
     * @param provenanceStore the URL of the provenance store to query
     * 
     * @return a DOM Document containing the results of the XQuery
     * @throws Exception
     */
    public Document query (String xquery, URL provenanceStore) throws Exception {
        StringBuffer sb = new StringBuffer ();
        sb.append ("<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">");
        
        sb.append ("<soapenv:Header>");
        sb.append ("</soapenv:Header>");
        sb.append ("<soapenv:Body>");
        sb.append ("<xq:query xmlns:xq = \"" + Constants.WSXQUERY_NS + "\">");
        sb.append ("<xq:xquery>");
        
        xquery = xquery.replaceAll ("<", "&lt;");
        xquery = xquery.replaceAll (">", "&gt;");
        
        sb.append (xquery);
        
        sb.append ("</xq:xquery>");
        sb.append ("</xq:query>");
        sb.append ("</soapenv:Body>");
        sb.append ("</soapenv:Envelope>");
        
        DefaultDocumentClient ddc = new DefaultDocumentClient ();
        
        return ddc.invoke (provenanceStore, sb.toString ());
    }

    /**
     * Send an XPath to the specified provenance store, specifying the form of the
     * document returned.  The root and item arguments specify the element names
     * at the top level of the response document.  For example, if the xpath is
     * "/ps:interactionRecord/ps:interactionKey", the root is "ex:root" and the
     * item is "ex:result", then the response takes the form:
     * <PRE>
     *   &lt;ex:root&gt;
     *     &lt;ex:result&gt;
     *       &lt;ps:interactionKey&gt; ...
     *     &lt;/ex:result&gt;
     *     &lt;ex:result&gt;
     *       &lt;ps:interactionKey&gt; ...
     *     &lt;/ex:result&gt;
     *     ...
     *   &lt;/ex:root&gt;
     * </PRE>
     * 
     * @param xpath           the XPath string
     * @param root            the qname of the root element of the response document
     * @param item            the qname of the element surrounding each result in the response document
     * @param namespaces      a mapping of prefix to namespace used in the XPath or for root/item
     * @param provenanceStore the URL of the provenance store to query
     * 
     * @return a DOM Document containing the results of the XQuery
     * @throws Exception
     */
    public Document xpath (String xpath, String root, String item, Map namespaces, URL provenanceStore) throws Exception {
        String xquery = "&lt;" + root + "&gt;" +
                " {for $n in $ps:pstruct" + xpath + " return" +
                " &lt;" + item + "&gt; {$n}" +
                " &lt;/" + item + "&gt;} " +
                " &lt;/" + root + "&gt;";
        Iterator eachPrefix = namespaces.keySet ().iterator ();
        String   prefix;
        String   namespace;
               
        while (eachPrefix.hasNext ()) {
            prefix    = (String) eachPrefix.next ();
            namespace = (String) namespaces.get (prefix);
            xquery = "declare namespace " + prefix + " = \"" + namespace + "\"; " + xquery;
        }
        
        return query (xquery, provenanceStore);
    }
    
    public Document provenanceQuery (String qdhXPath, String scopeXPath, Map namespaces, URL provenanceStore) throws Exception {
        ProvenanceQueryRequestDocument request;
        Document                       response;
        
        request  = new ProvenanceQueryRequestDocument (qdhXPath, scopeXPath, namespaces);
        response = new DefaultDocumentClient ().invoke (provenanceStore, request);
        
        return response;
    }
}
