/********************************************************************************
 * Created on Sep 16, 2005
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/

package org.pasoa.util.rdf;

import java.util.Iterator;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.pasoa.common.structure.ActorStatePAssertion;
import org.pasoa.common.structure.GlobalPAssertionKey;
import org.pasoa.common.structure.InteractionPAssertion;
import org.pasoa.common.structure.InteractionRecord;
import org.pasoa.common.structure.InteractionKey;
import org.pasoa.common.structure.Link;
import org.pasoa.common.structure.ObjectID;
import org.pasoa.common.structure.Relationship;
import org.pasoa.common.structure.SubjectID;
import org.pasoa.common.structure.View;
import org.pasoa.dommap.DOMWriter;
import org.pasoa.storage.BackendStore;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class RDFExporter {
    private final static String RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
    private final static String PSTRUCT_OWL = "http://www.pasoa.org/ontologies/pstructure#";
    private final static String STRING_XSD = "http://www.w3.org/2001/XMLSchema#string";
    
    public Document export (BackendStore bs) throws Exception {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance ();
        dbf.setNamespaceAware (true);
        DocumentBuilder docBuilder = dbf.newDocumentBuilder ();
        
        Document doc = docBuilder.newDocument ();
        
        Element top = doc.createElementNS (RDF_NS, "rdf");
        doc.appendChild (top);
        
        
        long numIRs = bs.getNumberOfInteractionRecords ();
        for (long i = 0; i < numIRs; i++) {
            InteractionRecord ir = bs.getInteractionRecord (i);
            writeInteractionRecord (ir, doc);
            
        }
        return doc;
    }
    
    public void writeInteractionRecord (InteractionRecord ir, Document doc) {
        String idObj          = writeIdentifiers (ir.getInteractionKey (), doc);
        String clientViewObj  = writeView (ir.getInteractionKey (), "client", ir.getSenderView (), doc);
        String serviceViewObj = writeView (ir.getInteractionKey (), "service", ir.getReceiverView (), doc);
        
        writeObjectProperty (doc,
                ir.getInteractionKey ().getInteractionID (),
                new QName (PSTRUCT_OWL, "interactionRecord"),
                new QName (PSTRUCT_OWL, "isIdentifiedBy"),
                idObj);
        
        writeObjectProperty (doc,
                ir.getInteractionKey ().getInteractionID (),
                new QName (PSTRUCT_OWL, "interactionRecord"),
                new QName (PSTRUCT_OWL, "hasClientView"),
                clientViewObj);
        
        writeObjectProperty (doc,
                ir.getInteractionKey ().getInteractionID (),
                new QName (PSTRUCT_OWL, "interactionRecord"),
                new QName (PSTRUCT_OWL, "hasServiceView"),
                serviceViewObj);
        
    }
    
    public String writeView (InteractionKey interactionRecordID, String viewName, View view, Document doc) {
        String   name = interactionRecordID.getInteractionID () + "/" + viewName;
        Iterator eachAssert;
        String   object;
        
        eachAssert = view.getInteractionPAssertions ().iterator ();
        while (eachAssert.hasNext ()) {
            object = writeInteractionPAssertion (name, (InteractionPAssertion) eachAssert.next (), doc);
            writeObjectProperty (doc, name, new QName (PSTRUCT_OWL, "view"), new QName (PSTRUCT_OWL, "hasInteractionPAssertion"), object);
        }
        eachAssert = view.getActorStatePAssertions ().iterator ();
        while (eachAssert.hasNext ()) {
            object = writeActorStatePAssertion (name, (ActorStatePAssertion) eachAssert.next (), doc);
            writeObjectProperty (doc, name, new QName (PSTRUCT_OWL, "view"), new QName (PSTRUCT_OWL, "hasActorStatePAssertion"), object);
        }
        eachAssert = view.getRelationships ().iterator ();
        while (eachAssert.hasNext ()) {
            object = writeRelationshipPAssertion (interactionRecordID, viewName, name, (Relationship) eachAssert.next (), doc);
            writeObjectProperty (doc, name, new QName (PSTRUCT_OWL, "view"), new QName (PSTRUCT_OWL, "hasRelationshipPAssertion"), object);
        }
        
        return name;
    }
    
    public String writeInteractionPAssertion (String prefix, InteractionPAssertion assertion, Document doc) {
        String name = prefix + "/" + assertion.getLocalPAssertionID ();
        
        writeDataProperty (doc, name, "interactionPAssertion", "hasLocalPAssertionID", STRING_XSD, assertion.getLocalPAssertionID ());
        writeDataProperty (doc, name, "interactionPAssertion", "asserts",              STRING_XSD, DOMWriter.writeToString (assertion.getMessageBody ()));
        writeDataProperty (doc, name, "interactionPAssertion", "hasDocumentStyle",     STRING_XSD, assertion.getDocumentationStyle ());
        
        return name;
    }
    
    public String writeActorStatePAssertion (String prefix, ActorStatePAssertion assertion, Document doc) {
        String name = prefix + "/" + assertion.getLocalPAssertionID ();
        
        writeDataProperty (doc, name, "actorStatePAssertion", "hasLocalPAssertionID", STRING_XSD, assertion.getLocalPAssertionID ());
        writeDataProperty (doc, name, "actorStatePAssertion", "asserts",              STRING_XSD, DOMWriter.writeToString (assertion.getActorStateBody ()));
        
        return name;
    }
    
    public String writeRelationshipPAssertion (InteractionKey interactionRecordID, String viewName, String prefix, Relationship assertion, Document doc) {
        String   name       = prefix + "/" + assertion.getLocalPAssertionID ();
        Iterator eachObject = assertion.getObjectIDs ().iterator ();
        String   object;
        int      count      = 1;
        
        writeDataProperty   (doc, name, "relationshipPAssertion", "hasLocalPAssertionID", STRING_XSD, assertion.getLocalPAssertionID ());
        writeObjectProperty (doc, name, "relationshipPAssertion", "hasSubject",    writeSubject (interactionRecordID, viewName, name, assertion.getRelation (), assertion.getSubject (), doc));
        while (eachObject.hasNext ()) {
            writeObjectProperty (doc, name, "relationshipPAssertion", "hasObject", writeObject (interactionRecordID, viewName, name, assertion.getRelation (), (ObjectID) eachObject.next (), count, doc));
            count += 1;
        }
        
        return name;
    }
    
    public String resourceName (GlobalPAssertionKey gpaid) {
        return gpaid.getInteractionKey ().getInteractionID () + "/" + gpaid.getViewName () + "/" + gpaid.getLocalPAssertionID ();
    }
    
    public String writeArgument (String name, String operation, GlobalPAssertionKey object, String parameterName, Element dataID, Document doc) {
        if (dataID == null) {
            writeObjectProperty (doc, name, "argument", "hasValue", resourceName (object));
        } else {
            writeObjectProperty (doc, name, "argument", "hasValue", writeDataItem (name, object, dataID, doc));
        }
        writeGlobalObjectProperty (doc, name, "argument", parameterName, operation);
        
        return name;
    }
    
    public String writeDataItem (String prefix, GlobalPAssertionKey object, Element dataID, Document doc) {
        String name = prefix + "/dataID";
        
        writeObjectProperty (doc, name, "dataItem", "hasLocation", writeElementValue (name, "dataID", dataID, doc));
        writeObjectProperty (doc, name, "dataItem", "isWithin",    resourceName (object));
        
        return name;
    }
    
    public String writeSubject (InteractionKey interactionRecordID, String viewName, String prefix, String relation, SubjectID subject, Document doc) {
        GlobalPAssertionKey gpaid = new GlobalPAssertionKey (interactionRecordID, viewName, subject.getLocalPAssertionID ());
        
        return writeArgument (prefix + "/subjectId", relation, gpaid, subject.getParameterName (), subject.getDataAccessor (), doc);
    }
    
    public String writeObject (InteractionKey interactionRecordID, String viewName, String prefix, String relation, ObjectID object, int index, Document doc) {
        GlobalPAssertionKey gpaid = object.getGlobalPAssertionKey ();
        
        return writeArgument (prefix + "/objectId" + index, relation, gpaid, object.getParameterName (), object.getObjectDataAccessor (), doc);
    }
    
    public String writeIdentifiers (InteractionKey irid, Document doc) {
        String interactionIdObj = writeInteractionId (irid.getInteractionID (), doc);
        String msgSourceObj = writeMsgSource (irid.getInteractionID (), irid.getMessageSource (), doc);
        String msgSinkObj = writeMsgSink (irid.getInteractionID (), irid.getMessageSink (), doc);
        
        writeObjectProperty (doc,
                irid.getInteractionID () + "/interactionKey",
                new QName (PSTRUCT_OWL, "interactionKey"),
                new QName (PSTRUCT_OWL, "includes"),
                interactionIdObj);
        
        writeObjectProperty (doc,
                irid.getInteractionID () + "/interactionKey",
                new QName (PSTRUCT_OWL, "interactionKey"),
                new QName (PSTRUCT_OWL, "includes"),
                msgSourceObj);
        
        writeObjectProperty (doc,
                irid.getInteractionID () + "/interactionKey",
                new QName (PSTRUCT_OWL, "interactionKey"),
                new QName (PSTRUCT_OWL, "includes"),
                msgSinkObj);
        
        return irid.getInteractionID () + "/interactionKey";
    }
    
    public String writeInteractionId (String interactionId, Document doc) {
        writeDataProperty (doc,
                interactionId + "/identifiers/interactionId",
                new QName (PSTRUCT_OWL, "interactionID"),
                new QName (RDF_NS, "value"),
                STRING_XSD,
                interactionId);
        
        return interactionId + "/identifiers/interactionId";
    }
    
    public String writeElementValue (String name, String type, Element value, Document doc) {
        writeDataProperty (doc, name, new QName (PSTRUCT_OWL, type), new QName (RDF_NS, "value"), STRING_XSD, DOMWriter.writeToString (value));
        return name;
    }
    
    public String writeClientId (String interactionId, Element clientId, Document doc) {
        return writeElementValue (interactionId + "/identifiers/clientId", "clientID", clientId, doc);
    }
    
    public String writeMsgSource (String interactionId, Element msgSource, Document doc) {
        return writeElementValue (interactionId + "/identifiers/msgSource", "messageSource", msgSource, doc);
    }
    
    public String writeServiceId (String interactionId, Element serviceid, Document doc) {
        return writeElementValue (interactionId + "/identifiers/serviceId", "serviceID", serviceid, doc);
    }
    
    public String writeMsgSink (String interactionId, Element msgSink, Document doc) {
        return writeElementValue (interactionId + "/identifiers/msgSink", "messageSink", msgSink, doc);
    }
    
    public void writeObjectProperty (Document doc, String subj, String subjType, String rel, String obj) {
        writeObjectProperty (doc, subj, new QName (PSTRUCT_OWL, subjType), new QName (PSTRUCT_OWL, rel), obj);
    }
    
    public void writeObjectProperty (Document doc,
            String subj,
            QName subjType,
            QName rel,
            String obj) {
        Element rdfDesc = doc.createElementNS (subjType.getNamespaceURI (), subjType.getLocalPart ());
        rdfDesc.setAttributeNS (RDF_NS, "id", subj);
        Element property = doc.createElementNS (rel.getNamespaceURI (), rel.getLocalPart ());
        property.setAttributeNS (RDF_NS, "resource", "#" + obj);
        rdfDesc.appendChild (property);
        doc.getDocumentElement ().appendChild (rdfDesc);
    }
    
    public void writeGlobalObjectProperty (Document doc, String subj, String subjType, String rel, String obj) {
        writeGlobalObjectProperty (doc, subj, new QName (PSTRUCT_OWL, subjType), new QName (PSTRUCT_OWL, rel), obj);
    }
    
    public void writeGlobalObjectProperty (Document doc, String subj, QName subjType, QName rel, String obj) {
        Element rdfDesc = doc.createElementNS (subjType.getNamespaceURI (), subjType.getLocalPart ());
        rdfDesc.setAttributeNS (RDF_NS, "id", subj);
        Element property = doc.createElementNS (rel.getNamespaceURI (), rel.getLocalPart ());
        property.setAttributeNS (RDF_NS, "resource", obj);
        rdfDesc.appendChild (property);
        doc.getDocumentElement ().appendChild (rdfDesc);
    }
    
    public void writeDataProperty (Document doc, String subj, String subjType, String rel, String objType, String obj) {
        writeDataProperty (doc, subj, new QName (PSTRUCT_OWL, subjType), new QName (PSTRUCT_OWL, rel), objType, obj);
    }
    
    public void writeDataProperty (Document doc,
            String subj,
            QName subjType,
            QName rel,
            String objType,
            String obj)
            
    {
        Element rdfDesc = doc.createElementNS (subjType.getNamespaceURI (), subjType.getLocalPart ());
        rdfDesc.setAttributeNS (RDF_NS, "id", subj);
        Element property = doc.createElementNS (rel.getNamespaceURI (), rel.getLocalPart ());
        property.setAttributeNS (RDF_NS, "datatype", objType);
        Node txtNode = doc.createTextNode (obj);
        property.appendChild (txtNode);
        rdfDesc.appendChild (property);
        doc.getDocumentElement ().appendChild (rdfDesc);
    }
}
